﻿// Solution created by Boyan Mihaylov
// http://code.dortikum.net

using System;
using System.Collections.Generic;
using System.ComponentModel.Composition;
using System.ComponentModel.Composition.Hosting;
using System.ComponentModel.Composition.Primitives;
using System.Linq;
using System.Reflection;

namespace BoyanMihaylov.Solution
{
	public sealed class FluentComposablePartDefinition : ComposablePartDefinition, IExport
	{
		private readonly Action<string, ComposablePartDefinition> _updatePartDefinition;
		private readonly Dictionary<string, object> _metadata;
		private readonly string _contractName;

		private IEnumerable<ExportDefinition> _exportDefinitions;
		private List<ImportDefinition> _importDefinitions;
		private Type _importType;
		private Type _exportType;

		public FluentComposablePartDefinition(Type type, string contractName, string typeIdentity, Action<string, ComposablePartDefinition> updatePartDefinitionCallback)
		{
			if (type == null)
				throw new ArgumentNullException("type");

			if (String.IsNullOrWhiteSpace(contractName))
				throw new ArgumentNullException("contractName");

			if (String.IsNullOrWhiteSpace(typeIdentity))
				throw new ArgumentNullException("typeIdentity");			

			if (updatePartDefinitionCallback == null)
				throw new ArgumentNullException("updatePartDefinitionCallback"); 

			_exportType = type;
			_contractName = contractName;
			_updatePartDefinition = updatePartDefinitionCallback;
			_importDefinitions = new List<ImportDefinition>();
			
			_metadata = new Dictionary<string, object>();
			_metadata.Add(CompositionConstants.ExportTypeIdentityMetadataName, typeIdentity);
		}

		public void AsSelf()
		{
			As(_exportType);
		}

		public void As(Type type)
		{
			if (!_exportType.IsAssignableFrom(type))
				throw new ArgumentOutOfRangeException(String.Format("Cannot cast from {0} to {1}", type.FullName, _exportType.FullName));

			if (type.IsInterface || type.IsAbstract)
				throw new ArgumentOutOfRangeException(String.Format("Cannot create instance of type {0}", type.FullName));

			_importType = type;

			// we take the first constructor only. if it has arguments, we assume
			// that they are going to be imported.
			ConstructorInfo ci = _importType.GetConstructors().FirstOrDefault();
			if (ci != null)
			{
				foreach (ParameterInfo pi in ci.GetParameters())
				{
					_importDefinitions.Add(new FluentImportDefinition(pi.ParameterType.GetContractName(), pi.ParameterType.GetTypeIdentity()));
				}
			}

			// add this partdefinition to the catalog
			_updatePartDefinition(_importType.GetContractName(), this);
		}

		public IExport WithPolicy(CreationPolicy policy)
		{
			// if the policy is Any, we do not put it in the metadata.
			if (policy == CreationPolicy.Any)
			{
				if (_metadata.ContainsKey(CompositionConstants.PartCreationPolicyMetadataName))
				{
					_metadata.Remove(CompositionConstants.PartCreationPolicyMetadataName);
				}
			}
			else
			{
				_metadata[CompositionConstants.PartCreationPolicyMetadataName] = policy;
			}
			return this;
		}

		public override IEnumerable<ExportDefinition> ExportDefinitions
		{
			get
			{
				if (_exportDefinitions == null)
				{
					_exportDefinitions = new ExportDefinition[] { new ExportDefinition(_contractName, _metadata) };
				}
				return _exportDefinitions;
			}
		}

		public override IEnumerable<ImportDefinition> ImportDefinitions
		{
			get { return _importDefinitions.ToArray(); }
		}

		public override IDictionary<string, object> Metadata
		{
			get { return _metadata; }
		}

		public override ComposablePart CreatePart()
		{
			return new FluentComposablePart(_importType, this);
		}

	}
}
